# _*_ coding: utf-8 _*_
"""
Time:     2021/12/22 11:35
Author:   ZHANG Yuwei
Version:  V 0.1
File:     __init__.py
Describe:
"""
from .main import SupertadSparse
from evaluation import Evaluate

name = 'SuperTLD'